#include "Image.hh"
#include "fft.hh"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

void Image_to_complex(Image<int>& p,COMPLEX **c)

{
  int index=0;
  int* PIX=p.PI();
  int L=p.PL();
  int W=p.PW();
  int i,j ; 
 
  for (j=0;j<L;j++)
    for (i=0;i<W;i++) {
	  c[j][i].real=(double)(PIX[index++]);
	  c[j][i].imag=0.0;
	}
}

void Complex_to_image(COMPLEX **c,Image<int>& p,double *min,double *max)

{
  int index=0;
  int* PIX=p.PI();
  int L=p.PL();
  int W=p.PW();
  int i,j;
  
  (*min)=255;
  (*max)=0;
  
  for (j=0;j<L;j++)
    for (i=0;i<W;i++) {
	  if (c[j][i].real<(*min)) (*min)=c[j][i].real;
	  else if (c[j][i].real>(*max)) (*max)=c[j][i].real;
	}
  for (j=0;j<L;j++)
    for (i=0;i<W;i++) {
	  PIX[index++]=(int)(round((c[j][i].real-(*min))*(255/((*max)-(*min)))));
	}
}

void Fourier_to_complex(Image<int>& module,Image<int>& phase,COMPLEX **c)

{
  int index=0;
  int* MODULE=module.PI();
  int* PHASE=phase.PI();
  int L=module.PL();
  int W=module.PW();
  int i,j ; 
 
  for (j=0;j<L;j++)
    for (i=0;i<W;i++) {
       c[i][j].real=MODULE[index]*cos((double)PHASE[index]);
	   c[i][j].imag=MODULE[index]*sin((double)PHASE[index]);;
	}
}

double SqModC(COMPLEX z) {
 return z.real*z.real + z.imag*z.imag;
}

int Powerof2(int n,int *m,int *twopm)
{
        if (n <= 1) {
                *m = 0;
                *twopm = 1;
                return(FALSE);
        }

   *m = 1;
   *twopm = 2;
   do {
      (*m)++;
      (*twopm) *= 2;
   } while (2*(*twopm) <= n);

   if (*twopm != n) 
                return(FALSE);
        else
                return(TRUE);

}



int FFT2D(COMPLEX **c,int nx,int ny,int dir)
{
   int i,j;
   int m,twopm;
   double *real,*imag;
   
  /* Transform the rows */
   real = (double *)malloc(nx * sizeof(double));
   imag = (double *)malloc(nx * sizeof(double));
   if (real == NULL || imag == NULL)
      return(FALSE);
   if (!Powerof2(nx,&m,&twopm) || twopm != nx)
	// power(2,nx);
      return(FALSE);
   for (j=0;j<ny;j++) {
      for (i=0;i<nx;i++) {
         real[i] = c[i][j].real;
         imag[i] = c[i][j].imag;
      }
      FFT(dir,m,real,imag);
      for (i=0;i<nx;i++) {
         c[i][j].real = real[i];
         c[i][j].imag = imag[i];
      }
   }
   free(real);
   free(imag);

   /* Transform the columns */
   real = (double *)malloc(ny * sizeof(double));
   imag = (double *)malloc(ny * sizeof(double));
   if (real == NULL || imag == NULL)
      return(FALSE);
  
   if (!Powerof2(ny,&m,&twopm) || twopm != ny)
      return(FALSE);
   for (i=0;i<nx;i++) {
      for (j=0;j<ny;j++) {
         real[j] = c[i][j].real;
         imag[j] = c[i][j].imag;
      }
      FFT(dir,m,real,imag);
      for (j=0;j<ny;j++) {
         c[i][j].real = real[j];
         c[i][j].imag = imag[j];
      }
   }
   free(real);
   free(imag);

   return(TRUE);
}


int FFT(int dir,int m,double *x,double *y)
{
   long nn,i,i1,j,k,i2,l,l1,l2;
   double c1,c2,tx,ty,t1,t2,u1,u2,z;
   
   /* Calculate the number of points */
   nn = 1;
   for (i=0;i<m;i++)
      nn *= 2;

   /* Do the bit reversal */
   i2 = nn >> 1;
   j = 0;
   for (i=0;i<nn-1;i++) {
      if (i < j) {
         tx = x[i];
         ty = y[i];
         x[i] = x[j];
         y[i] = y[j];
         x[j] = tx;
         y[j] = ty;
      }
      k = i2;
      while (k <= j) {
         j -= k;
         k >>= 1;
      }
      j += k;
   }

   /* Compute the FFT */
   c1 = -1.0;
   c2 = 0.0;
   l2 = 1;
   for (l=0;l<m;l++) {
      l1 = l2;
      l2 <<= 1;
      u1 = 1.0;
      u2 = 0.0;
      for (j=0;j<l1;j++) {
         for (i=j;i<nn;i+=l2) {
            i1 = i + l1;
            t1 = u1 * x[i1] - u2 * y[i1];
            t2 = u1 * y[i1] + u2 * x[i1];
            x[i1] = x[i] - t1;
            y[i1] = y[i] - t2;
            x[i] += t1;
            y[i] += t2;
         }
         z =  u1 * c1 - u2 * c2;
         u2 = u1 * c2 + u2 * c1;
         u1 = z;
      }
      c2 = sqrt((1.0 - c1) / 2.0);
      if (dir == 1)
         c2 = -c2;
      c1 = sqrt((1.0 + c1) / 2.0);
   }

  //  /* Scaling for forward transform */
   if (dir == 1) {
      for (i=0;i<nn;i++) {
         x[i] /= (double)nn;
         y[i] /= (double)nn;
      }
   }
   return(TRUE);
}
 
void Calcul_Module(Image<int>& p,COMPLEX  **c,double *min,double *max)
{
  int indexbis,index=0;
  double mod,val;
  double *MODULE;
  int* PIX=p.PI();
  int L=p.PL();
  int W=p.PW();
  int i,j,a,b; 
  double valmin = -1000;
  
  (*max)=0;
  (*min)=0;
  MODULE = (double*)calloc(W*L,sizeof(double));
  for (j=0;j<L;j++)
    for (i=0;i<W;i++) {
	  mod = sqrt((c[i][j].real*c[i][j].real)+(c[i][j].imag*c[i][j].imag));
	  if (mod) {
		val = log(sqrt((c[i][j].real*c[i][j].real)+(c[i][j].imag*c[i][j].imag)));
		if ((*max) < val) (*max)=val;
		if ((*min) > val) (*min)=val;
		MODULE[index]=val;
	  } else
		MODULE[index]=valmin;
	  index++;
	}
  printf("min = %f, max = %f\n",(*min),(*max));
  index=0;
  for (j=0;j<L;j++)
	for (i=0;i<W;i++) {
	  a = i+W/2;
	  b = j+L/2;
	 if (a>=W) a = i-W/2;
	  if (b>=L) b = j-L/2;
	  indexbis = a*W+b;
	  if (MODULE[index] == valmin) PIX[indexbis] = 0;
	  else PIX[indexbis]=(int)(round((MODULE[index]-(*min))*(255/((*max)-(*min)))));
	  index++;
	}
  free(MODULE);
}

void Calcul_phase(Image<int>& p,COMPLEX  **c)

{
  int index=0;
  Image<int> q(p);
  int* PIX=p.PI();
  int L=p.PL();
  int W=p.PW();
  int i,j,a,b; 
  double val;
  double PI=3.141592635;
  
  for (int j=0;j<L;j++)
    for (int i=0;i<W;i++) {
      a = i+W/2;
	  b = j+L/2;
	  if (a>=W) a = i-W/2;
	  if (b>=L) b = j-L/2;
	  val = atan2(c[i][j].imag,c[i][j].real);
	  PIX[a*W+b]=(int)(round((val+PI)*(255/(2*PI))));
	  index++;
	}
}

void Haut_Bas(COMPLEX **c,int w,int h,int frequence,int type,int forme) {
  int i,j;
  int f1,f2,f3;

  f1 = w-frequence;
  f2 = h-frequence;
  f3 = frequence*frequence;
  if (forme==0) {
    if (type==0) {
      for (j=0;j<h;j++)
        for (i=0;i<w;i++) {
          if (((i>=frequence)&&(i<f1))||
              ((j>=frequence)&&(j<f2)))
            {c[i][j].real=0;c[i][j].imag=0;}
        }
    } else {
      for (j=0;j<h;j++)
        for (i=0;i<w;i++) {
          if (((i<frequence)||(i>=f1))&&
              ((j<frequence)||(j>=f2)))
            {c[i][j].real=0;c[i][j].imag=0;}
        }
    }
  } else {
    if (type==0) {
      for (j=0;j<h;j++)
        for (i=0;i<w;i++) {
          if ((((i-w)*(i-w))+((j-h)*(j-h)) > f3) &&
              ((i*i+j*j) > f3) &&
              (((i*i)+(j-h)*(j-h)) > f3) &&
              ((((i-w)*(i-w))+(j*j)) > f3))
            {c[i][j].real=0;c[i][j].imag=0;}
        }
    } else {
      for (j=0;j<h;j++)
    for (i=0;i<w;i++) {
      if ((((i-w)*(i-w))+((j-h)*(j-h)) < f3) ||
          ((i*i+j*j) < f3) ||
          (((i*i)+(j-h)*(j-h)) < f3) ||
          ((((i-w)*(i-w))+(j*j)) < f3))
        {c[i][j].real=0;c[i][j].imag=0;}
        }
    }
  }
}

void Passe_Coupe(COMPLEX  **c,int w,int h,int freq_min,int freq_max,int type,int
 forme)
{
  int i,j;
  int f1,f2,f3,g1,g2,g3;

  f1 = w-freq_min;g1 = w-freq_max;
  f2 = h-freq_min;g2 = h-freq_max;
  f3 = freq_min*freq_min;g3 = freq_max*freq_max;
  if (forme==0) {
    if (type==0) {
      for (j=0;j<h;j++)
        for (i=0;i<w;i++) {
          if ((((i>=freq_max)&&(i<g1))||((j>=freq_max)&&(j<g2))) ||
              (((i<freq_min)||(i>=f1))&&((j<freq_min)||(j>=f2))))
            {c[i][j].real=0;c[i][j].imag=0;}
        }
    } else {
      for (j=0;j<h;j++)
        for (i=0;i<w;i++) {
          if ((((i<freq_max)||(i>=g1))&&((j<freq_max)||(j>=g2))) &&
              (((i>=freq_min)&&(i<f1))||((j>=freq_min)&&(j<f2))))
            {c[i][j].real=0;c[i][j].imag=0;}
        }
    }
  } else {
    if (type==0) {
      for (j=0;j<h;j++)
        for (i=0;i<w;i++) {
          if (((((i-w)*(i-w))+((j-h)*(j-h)) > g3) && ((i*i+j*j) > g3) &&
               (((i*i)+(j-h)*(j-h)) > g3) && ((((i-w)*(i-w))+(j*j)) > g3)) ||
              ((((i-w)*(i-w))+((j-h)*(j-h)) < f3) ||
               (((i*i+j*j) < f3) || (((i*i)+(j-h)*(j-h)) < f3) ||
                ((((i-w)*(i-w))+(j*j)) < f3))))
            {c[i][j].real=0;c[i][j].imag=0;}
        }
    } else {
      for (j=0;j<h;j++)
        for (i=0;i<w;i++) {
          if (((((i-w)*(i-w))+((j-h)*(j-h)) < g3) || ((i*i+j*j) < g3) ||
               (((i*i)+(j-h)*(j-h)) < g3) || ((((i-w)*(i-w))+(j*j)) < g3)) &&
              ((((i-w)*(i-w))+((j-h)*(j-h)) > f3) && (((i*i+j*j) > f3) &&
               (((i*i)+(j-h)*(j-h)) > f3) && ((((i-w)*(i-w))+(j*j)) > f3))))
                {c[i][j].real=0;c[i][j].imag=0;}
        }
    }
  }
}

void Ghost(COMPLEX **c,int w,int h,int composante,int type) {
  int i,j;
  double PI=3.141592635;
  double module,phase;
  double r1,r2,x;
  
  if (composante == 0) {// Meme amplitude, Phase alatoire
      for (j=0;j<h;j++)
	for (i=0;i<w;i++) {
	  module = sqrt(c[i][j].real*c[i][j].real+c[i][j].imag*c[i][j].imag);
	  phase = (rand()/(double)RAND_MAX)* 2*PI - PI;
	  c[i][j].real = module*cos(phase);
	  c[i][j].imag = module*sin(phase);
	}
    } else {// Meme phase, Amplitude constante
    for (j=0;j<h;j++)
      for (i=0;i<w;i++) {
	phase = atan2(c[i][j].imag,c[i][j].real);
	module = 255;
	c[i][j].real = module*cos(phase);
	c[i][j].imag = module*sin(phase);
      }
  }
}

void TH_Spectrum(COMPLEX **c,int w,int h,float seuil_log) {
  int i,j;
  int index,cpt;
  double *sqmod;
  double sqmax;
  float seuil_relatif = pow(10,-seuil_log);
   
  printf("Seuil Relatif de l'amplitude Max : %f\n",seuil_relatif);
  // Calcul de l'image d'amplitude au carree
  sqmod = (double*)calloc(w*h,sizeof(double));
  index = 0;
  sqmax = 0.0;
  for (j=0;j<h;j++)
    for (i=0;i<w;i++) {
       sqmod[index] = SqModC(c[i][j]);
       if ((sqmod[index] > sqmax)&&(index !=0)) sqmax = sqmod[index];
       index++;
      }
  index=0;
  cpt=0;
  for (j=0;j<h;j++)
    for (i=0;i<w;i++) {
      if (SqModC(c[i][j]) < seuil_relatif*sqmax) {
        c[i][j].real = 0;
        c[i][j].imag = 0;
        cpt++;
        }
    index++;
    }
  printf("Pourcentage de Frequences utiles : %0.3f %%\n",100.0*(1 - (float)(cpt)/(w*h)));
  free(sqmod);
}

void NMS_Spectrum(COMPLEX **c,int w,int h,int rayon_ex,int nb_maxima) {
  int i,j,k;
  int index;
  double *sqmod;
  int *INDEX_NMS;
  int *MSK;

  INDEX_NMS = (int*)calloc(nb_maxima,sizeof(int));
  // Calcul de l'image d'amplitude au carree
  sqmod = (double*)calloc(w*h,sizeof(double));
  index = 0;
  for (j=0;j<h;j++)
    for (i=0;i<w;i++) 
       sqmod[index++] = c[i][j].real*c[i][j].real + c[i][j].imag*c[i][j].imag;
  // On calcule les nb_maxima max locaux, distants de rayon_ex
  Calcul_N_MAX(INDEX_NMS,sqmod,w,h,rayon_ex,nb_maxima);
  // Creation du masque 
  MSK = (int*)calloc(w*h,sizeof(int));
  for (k = 0;k < nb_maxima;k++) MSK[INDEX_NMS[k]] = 1;
  index=0;
  for (j=0;j<h;j++)
    for (i=0;i<w;i++) {
      if (MSK[index] == 0) {
        c[i][j].real = 0;
        c[i][j].imag = 0;
        }
    index++;
    }
  free(sqmod);free(INDEX_NMS);free(MSK);
}

// The following function outputs (in the array POSITION) indexes (positions) of the nb_max 
// largest values in the array TAB (sized width x height), with the condition that 
// 2 positions cannot be at a closer distance than rayon_ex
 
void Calcul_N_MAX(int *POSITION,double *TAB,int width,int height,int rayon_ex,int nb_max) {
int index,rang,i,j;
int taille;
double new_val;
double *VAL_MAX = (double *)calloc(nb_max,sizeof(double));
double *TAB_DILATE1 = (double *)calloc(width*height,sizeof(double));
double *TAB_DILATE2 = (double *)calloc(width*height,sizeof(double));
double *TAB_DILATE = (double *)calloc(width*height,sizeof(double));

double epsilon = 1.0e-60;//A utiliser en fonction des oprations ralise sur les double

//Prliminaire : on calcule la dilatation morphologique de la 
//valeur absolue de l'image TAB par un carr de ct taille = 2*rayon_ex + 1

taille = 2*rayon_ex + 1;

//Dilatation horizontale (mthode de Van Herk)
index = 0;// Phases 1 et 2 : dilatation bloc-rcursive
for (j = 0;j < height;j++) {
  for (i = 0;i < width;i++) {
    if (i % taille == 0) TAB_DILATE1[index] = fabs(TAB[index]);
      else TAB_DILATE1[index] = fmax(fabs(TAB[index]),TAB_DILATE1[index-1]);
    index++;
  }
  for (i = width-1;i >= 0;i--) {
    index--;
    if ((i % taille == 0)||(i == width-1)) TAB_DILATE2[index] = fabs(TAB[index]);
      else TAB_DILATE2[index] = fmax(fabs(TAB[index]),TAB_DILATE2[index+1]);
  }
 index += width;
}
index = 0;// Phase 3 : Combinaison des max
for (j = 0;j < height;j++) {
  for (i = 0;i < taille/2;i++) {
    TAB_DILATE[index] = fmax(TAB_DILATE1[index + taille/2],TAB_DILATE2[j*width]);
    index++;
    }
  for (i = taille/2;i < width-taille/2;i++) {
    TAB_DILATE[index] = fmax(TAB_DILATE1[index + taille/2],TAB_DILATE2[index - taille/2]);
    index++;
  }
  for (i = width-taille/2;i < width;i++) {
    TAB_DILATE[index] = fmax(TAB_DILATE1[(j+1)*width-1],TAB_DILATE2[index - taille/2]);
    index++;
  }
}
//Dilatation verticale (mthode de Van Herk)
// Phases 4 et 5 : dilatation bloc-rcursive
for (i = 0;i < width;i++) {
  index = i;
  for (j = 0;j < height;j++) {
    if (j % taille == 0) TAB_DILATE1[index] = TAB_DILATE[index];
      else TAB_DILATE1[index] = fmax(TAB_DILATE[index],TAB_DILATE1[index-width]);
    index += width;
  }
  for (j = height-1;j >= 0;j--) {
    index -= width;
    if ((j % taille == 0)||(j == height-1)) TAB_DILATE2[index] = TAB_DILATE[index];
      else TAB_DILATE2[index] = fmax(TAB_DILATE[index],TAB_DILATE2[index+width]);
  }
}
// Phase 6 : Combinaison des max
for (i = 0;i < width;i++) {
  index = i;
  for (j = 0;j < taille/2;j++) {
    TAB_DILATE[index] = fmax(TAB_DILATE1[index + width*(taille/2)],TAB_DILATE2[i]);
    index += width;
    }
  for (j = taille/2;j < height-taille/2;j++) {
    TAB_DILATE[index] = fmax(TAB_DILATE1[index + width*(taille/2)],TAB_DILATE2[index - width*(taille/2)]);
    index += width;
  }
  for (j = height-taille/2;j < height;j++) {
    TAB_DILATE[index] = fmax(TAB_DILATE1[width*(height-1)+i],TAB_DILATE2[index - width*(taille/2)]);
    index += width;
  }
}
for (index = 0;index < width*height;index++) {
    new_val = fabs(TAB[index]);
    if ((new_val > VAL_MAX[nb_max - 1])&&
          (TAB_DILATE[index] == new_val)) {//Attention condition d'galit fragile sur double
     rang = nb_max - 1;
     do {
        if (rang > 0) { 
           VAL_MAX[rang] = VAL_MAX[rang-1];
           POSITION[rang] = POSITION[rang-1];
	   rang--;
          }	
        } while ((new_val > VAL_MAX[rang]) && (rang > 0));
     VAL_MAX[rang] = new_val;//Insertion
     POSITION[rang] = index;  
   }      
}
free(VAL_MAX);
free(TAB_DILATE1);
free(TAB_DILATE2);
free(TAB_DILATE);
}
